import json
from typing import Dict, Set, Union

Interfaces = Dict[str, Dict[str, Union[str, int]]]


def read_interfaces(path: str) -> Interfaces:
	"""Reads an Interfaces JSON file at the given path and returns it as a dictionary."""
	with open(path, encoding='utf-8') as f:
		return json.load(f)


def factors(number: int, limit: int) -> Set[int]:
	return {x for x in [2**x for x in range(limit)] if number % x == 0}


def multiple(layer: int, limit: int) -> Set[str]:
	"""Returns a set of strings to be used as Slots with Pabianas default Clock.

	Args:
		layer: The layer in the hierarchy this Area is placed in.
			Technically, the number specifies how many of the Clocks signals are relevant to the Area.
			Between 1 and limit.
		limit: The number of layers of the hierarchy.
	"""
	return {str(x).zfill(2) for x in [2**x for x in range(limit)] if x % 2**(layer - 1) == 0}



import json
import logging
from typing import Any, Callable, Dict, Optional, Sequence, Set, Iterable

import zmq

from .extensions import Publisher, Pusher
from .node import Node
from .utils import caller
from .. import abcs
from ..parsers import init_full, imprint_full
from ..utils import Interfaces


logger = logging.getLogger(__name__)


class Area(Node, abcs.Area):
	"""Class that implements the Pabiana Area interface based on the ZMQ Node class.

	A ZMQ Pabiana Area can provide one Puller (as Receiver) and multiple Subscriber Interfaces.
	Other ZMQ Pabiana Areas could Push Triggers on and Publish messages to these Interfaces.
	"""
	def __init__(self, name: str, interfaces: Interfaces):
		super().__init__(name, interfaces)
		self._schedule_function = caller  # type: Callable

		self._triggers = {'exit': self.stop}  # type: Dict[str, Callable]
		self._demand = {}  # type: Dict[Callable, Dict[str, Any]]
		self._loop = {}  # type: Dict[Callable, Dict[str, Any]]

		self._processors = {}  # type: Dict[Optional[str], Dict[Optional[str], Callable]]
		self._alterations = set()  # type: Set[Callable]
		self._altered = set()  # type: Set[Callable]
		
		self._active_function = None  # type: Callable
		self._pulse_function = None  # type: Callable

		self.publish = Publisher(*self.rslv('pub'), self._zmq).publish
		self.trigger = Pusher(self, self._zmq).trigger

	def publish(self, message: dict, slot: str = None): pass

	def trigger(self, target: str, trigger: str, parameters: Dict[str, Any]={}): pass

	def scheduling(self, func: Callable) -> Callable:
		def combined(*args, **kwargs):
			caller(*func(*args, **kwargs))
		self._schedule_function = combined
		return func
	
	def subscribe(self, clock_name: str=None, clock_slots: Iterable[str]=None, subscriptions: Dict[str, Any]={}):
		"""Subscribes this Area to the given Areas and optionally given Slots. Must be called before the Area is run.

		Args:
			clock_name: The name of the Area that is used as synchronizing Clock.
			clock_slots: The slots of the Clock relevant to this Area.
			subscriptions: A dictionary containing the relevant Areas names as keys and optionally the Slots as values.
		"""
		for area in subscriptions:  # type: str
			init_full(self, area, subscriptions[area])
			subscriptions[area] = {'slots': subscriptions[area]}

		if clock_name is not None:
			self.clock_name = clock_name
			self.clock_slots = clock_slots
			subscriptions[clock_name] = {'slots': clock_slots, 'buffer-length': 1}
		self.setup(puller=True, subscriptions=subscriptions)
	
	def _process(self, interface: int, message: Sequence[str], source: str=None):
		if interface == zmq.PULL:
			message = json.loads(message[0])
			trigger_name = message['trigger']
			del message['trigger']
			self.comply(trigger_name, message)
		else:
			if source == self.clock_name:
				logger.log(5, 'Tick')
				self.proceed()
			else:
				slot = message[0]
				message = json.loads(message[1])
				self.process(source, message, slot)
	
	# ------------- Trigger processing functions -------------
	
	def register(self, func: Callable) -> Callable:
		self._triggers[func.__name__] = func
		return func
	
	def comply(self, trigger: str, parameters: Dict[str, Any]={}):
		try:
			logger.debug('Trigger call: "%s"', trigger)
			func = self._triggers[trigger]
			self._demand[func] = parameters
		except KeyError:
			logger.warning('Unavailable Trigger called')
		
	def autoloop(self, trigger: str, parameters: Dict[str, Any]={}):
		func = self._triggers[trigger]
		self._loop[func] = parameters
	
	# ------------- Message processing functions -------------
	
	def alteration(self, func: Callable=None, source: str=None, slot: str=None) -> Callable:
		def decorator(internal_func):
			if source not in self._processors:
				self._processors[source] = {}
			self._processors[source][slot] = internal_func
			return func
		if func:
			return decorator(func)
		return decorator
	
	def process(self, source: str, message: Dict, slot: str=None):
		try:
			imprint_full(self, source, message, slot)
			func = None
			if None in self._processors:
				func = self._processors[None][None]
			if source in self._processors:
				if slot in self._processors[source]:
					func = self._processors[source][slot]
				elif None in self._processors[source]:
					func = self._processors[source][None]
			if func is not None:
				self._alterations.add(func)
			logger.debug('Subscriber Message from "%s" - "%s"', source, slot)
		except KeyError:
			logger.debug('Unsubscribed Message from "%s" - "%s"', source, slot)
		
	def alter(self, source: str=None, slot: str=None):
		self._altered.add(self._processors[source][slot])
	
	# ------------- Clock processing functions -------------
	
	def activity(self, func: Callable) -> Callable:
		self._active_function = func
		return func
	
	def pulse(self, func: Callable) -> Callable:
		self._pulse_function = func
		return func
	
	def proceed(self):
		looped = None
		if self._loop:
			self._demand.update(self._loop)
			looped = set(self._loop)
			self._loop.clear()
		altered = None
		if self._altered:
			self._alterations.update(self._altered)
			altered = self._altered.copy()
			self._altered.clear()
		if self._demand or self._alterations:
			self._schedule_function(self._demand, self._alterations, looped=looped, altered=altered)
			if self._active_function is not None:
				self._active_function()
			self._demand.clear()
			self._alterations.clear()
		if self._pulse_function is not None:
			self._pulse_function()
		self.time += 1



from typing import Any, Callable, Iterable, Dict, List, Union

import zmq

Socket = zmq.sugar.socket.Socket
Context = zmq.sugar.context.Context
Poller = zmq.sugar.poll.Poller
Frame = zmq.sugar.frame.Frame


def decoder(rcvd: Iterable[bytes]) -> List[str]:
	result = [x.decode('utf-8') for x in rcvd]
	if result[0] == '':
		result[0] = None
	return result


def caller(*args: Union[Dict[Callable, Dict[str, Any]], Iterable[Callable]], **kwargs):
	for functions in args:
		if isinstance(functions, Dict):
			for func in functions:
				func(**functions[func])
		else:
			for func in functions:
				func()



from .parsers import init_full
from .utils import Interfaces, factors
from .zmqs.area import Area


def create_director(name: str, interfaces: Interfaces, limit: int=6):
	def aggregate(result, areas, keys):
		for key in keys:
			result += areas[key]
		return result

	def fill(result, size):
		for x in range(1, size+1):
			result[x] = set()
		return result

	area = Area(name, interfaces)
	area.subscribe()
	area.context.update({'areas': set(), 'mapping': fill({}, 32), 'ready': set(), 'factors': set(), 'target': set()})

	empty = {}

	@area.pulse
	def routine():
		area.publish(empty, slot=str(max(area.context['factors'])))

	process = area.process
	comply = area.comply

	def replace_process(source, message, slot=None):
		if source in area.context['areas']:
			if slot == 'status':
				area.context['ready'].add(source)
		process(source, message, slot)

		if area.context['target'].issubset(area.context['ready']):
			area.proceed()
			area.context['ready'] -= area.context['target']
			area.context['factors'] = factors(area.time, limit)
			area.context['target'] = aggregate(area.context['target'], area.context['mapping'], area.context['factors'])

	def replace_comply(trigger, parameters={}):
		if trigger == 'add':
			try:
				target = parameters['name']
				slot = int(parameters['slot'])
				area.add(target, slots=['status'])
				init_full(area, target, ['status'])
				area.context['areas'].add(target)
				area.context['mapping'][slot].add(target)
			except: pass
		comply(trigger, parameters)

	area.process = replace_process
	area.comply = replace_comply
	return area


def create_clock(name: str, interfaces: Interfaces, limit: int=6):
	area = Area(name, interfaces)
	area.subscribe()
	empty = {}

	@area.pulse
	def routine():
		area.publish(empty, slot=str(max(factors(area.time, limit))).zfill(2))

	area.routine = area.proceed
	return area



import logging

from .zmqs.area import Area
from .zmqs.node import Node
from .zmqs.runner import Runner
from .zmqs.extensions import Publisher, Pusher


logging.getLogger(__name__).addHandler(logging.NullHandler())

repo = {}



from collections import deque
from typing import Iterable, MutableMapping

from pabiana.abcs import Area


def init_full(area: Area, source: str, slots: Iterable[str]=None):
	if slots is None:
		area.context[source] = deque(maxlen=100)
	else:
		area.context[source] = {}
		for slot in slots:
			area.context[source][slot] = deque(maxlen=100)


def imprint_full(area: Area, source: str, message: MutableMapping, slot=None):
	if slot is None:
		area.context[source].appendleft(message)
		area.context[source][0]['time-rcvd'] = area.time
	else:
		area.context[source][slot].appendleft(message)
		area.context[source][slot][0]['time-rcvd'] = area.time


import logging
import signal
from abc import abstractmethod
from typing import Any, Callable, Dict, Iterable, Sequence

import zmq

from .utils import Context, decoder, Poller, Socket
from .. import abcs
from ..utils import Interfaces


logger = logging.getLogger(__name__)


class Node(abcs.Node):
	"""Abstract class that implements the Pabiana Node interface based on the ZMQ library.

	A ZMQ Pabiana Node can provide one Puller and multiple Subscriber Interfaces.
	Other ZMQ Pabiana Nodes could Push and Publish messages to these Interfaces.
	"""
	def __init__(self, name: str, interfaces: Interfaces):
		super().__init__(name, interfaces)
		self.routine = None  # type: Callable
		self._goon = None  # type: bool
		self._zmq = zmq.Context.instance()  # type: Context
		self._poller = zmq.Poller()  # type: Poller

		self._subscribers = {}  # type: Dict[Socket, str]
	
	def setup(self, puller: bool=None, subscriptions: Dict[str, Any]={}):
		"""Sets up this Node with the specified Interfaces before it is run.

		Args:
			puller: Indication if a Puller Interface should be created.
			subscriptions: Collection of the Subscriber Interfaces to be created and their Slots.
		"""
		if puller:
			puller = self._zmq.socket(zmq.PULL)
			ip, port, host = self.rslv('rcv')
			puller.bind('tcp://{}:{}'.format(host or ip, port))
			self.poll(puller)
		if subscriptions:
			for publisher in subscriptions:  # type: str
				self.add(publisher, subscriptions[publisher].get('slots'), subscriptions[publisher].get('buffer-length'))
			logger.info('Listening to %s', {
				k: (1 if subscriptions[k].get('slots') is None else len(subscriptions[k].get('slots')))
				for k in subscriptions
			})

	def add(self, publisher: str, slots: Iterable[str]=None, buffer_length=None):
		subscriber = self._zmq.socket(zmq.SUB)  # type: Socket
		# if buffer_length is not None:
		# 	subscriber.set_hwm(buffer_length)
		ip, port, host = self.rslv(name=publisher, interface='pub')
		subscriber.connect('tcp://{}:{}'.format(ip, port))
		if slots is None:
			slots = ['']
		for slot in slots:  # type: str
			subscriber.subscribe(slot)
		self._subscribers[subscriber] = publisher
		self.poll(subscriber)

	def poll(self, socket: Socket):
		self._poller.register(socket, zmq.POLLIN)
	
	def run(self, timeout: int=1000, linger: int=1000):
		self._goon = True
		signal.signal(signal.SIGINT, self.stop)
		try:
			while self._goon:
				socks = self._poller.poll(timeout)
				for sock, event in socks:
						self._process(sock.socket_type, decoder(sock.recv_multipart()), self._subscribers.get(sock))
				if self.routine is not None:
					self.routine()
		finally:
			self._zmq.destroy(linger=linger)
			logger.debug('Context destroyed')
			logging.shutdown()

	@abstractmethod
	def _process(self, interface: int, message: Sequence[str], source: str = None):
		pass
	
	def stop(self, *args, **kwargs):
		self._goon = False



import argparse
import importlib
import os
from os import path
from pip._internal import main as pip_main


from . import repo
from .templates import create_clock
from .utils import read_interfaces


def main():
	parser = argparse.ArgumentParser(description='Run a Pabiana area.')
	parser.add_argument('names', help='specify module and area name in the form of module_name:area_name')
	parser.add_argument('-f', '--fast', action='store_true', help='prevent the use of Pip before running the area.')
	arguments = parser.parse_args()
	module_name, area_name = arguments.names.split(':')

	base_path = os.getcwd()
	module_path = path.join(base_path, module_name)
	interfaces_path = path.join(base_path, 'interfaces.json')
	requirements_path = path.join(base_path, module_name, 'requirements.txt')

	if not arguments.fast and path.isfile(requirements_path):
		pip_main(['install', '-U', '-r', requirements_path])

	run(module_name, area_name, base_path, module_path, interfaces_path)


def run(module_name, area_name, base_path, module_path, interfaces_path):
	repo['module-name'] = module_name
	repo['area-name'] = area_name
	repo['interfaces'] = read_interfaces(interfaces_path)
	repo['base-path'] = base_path
	repo['module-path'] = module_path

	try:
		mod = importlib.import_module(module_name)
	except ImportError:
		if module_name == 'def_clock' and area_name == 'clock':
			area = create_clock(name=area_name, interfaces=repo['interfaces'])
			area.run(timeout=500)
			return
		else:
			raise

	if hasattr(mod, 'premise'):
		mod.premise()

	if hasattr(mod, 'area'):
		if hasattr(mod, 'config'):
			params = {
				'clock_name': mod.config.get('clock-name'),
				'clock_slots': mod.config.get('clock-slots')
			}
			if 'subscriptions' in mod.config:
				params['subscriptions'] = mod.config['subscriptions']
			mod.area.subscribe(**params)
			if 'context-values' in mod.config:
				mod.area.context.update(mod.config['context-values'])
		mod.area.run()

	elif hasattr(mod, 'runner'):
		if hasattr(mod.runner, 'setup'):
			params = {}
			if hasattr(mod, 'config'):
				params.update(mod.config)
			mod.runner.setup(**params)
		mod.runner.run()


if __name__ == '__main__':
	main()


from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Dict, Tuple, Set

from .utils import Interfaces


class Node(ABC):
	"""Abstract interface implemented by every Pabiana Node.
	
	A Pabiana Node is a type of networking software that can provide different types of Interfaces.
	Other Nodes can connect to these Interfaces.
	
	Attributes:
		name: Unique name of the Node.
		interfaces: IP addresses, ports and optionally host IPs for every Nodes interfaces are stored here.
		time: Internal time of the Node.
	"""
	def __init__(self, name: str, interfaces: Interfaces):
		self.name = name  # type: str
		self.interfaces = interfaces  # type: Interfaces
		self.time = 1  # type: int
	
	def rslv(self, interface: str, name: str=None) -> Tuple[str, int, Optional[str]]:
		"""Return the IP address, port and optionally host IP for one of this Nodes interfaces."""
		if name is None:
			name = self.name
		key = '{}-{}'.format(name, interface)
		host = self.interfaces[key].get('host', None)
		return self.interfaces[key]['ip'], self.interfaces[key]['port'], host
	
	@abstractmethod
	def run(self):
		"""Starts the Node. Runs in an infinite loop."""
		pass
	
	@abstractmethod
	def stop(self):
		"""Stops the Node safely."""
		pass


class Area(Node):
	"""Abstract interface implemented by every Pabiana Area.
	
	A Pabiana Area is a special type of Node.
	It can provide one Receiver and multiple Subscriber Interfaces.
	Other Areas could call remote Triggers on and Publish messages to these Interfaces.
	It can process Triggers and messages. It includes time synch mechanisms.

	Attributes:
		clock_name: Unique name of the Clock.
		clock_slot: Name of the publishing Slot of the Clock signal.
		context: Central store of the Areas non-volatile state.
	"""
	def __init__(self, name: str, interfaces: Interfaces):
		super().__init__(name, interfaces)
		self.clock_name = None  # type: str
		self.clock_slots = None  # type: Set[str]
		self.context = {}  # type: dict

	@abstractmethod
	def publish(self, message: dict, slot: str=None):
		"""Publishes a message on the optionally given Slot."""
		pass

	@abstractmethod
	def trigger(self, target: str, trigger: str, parameters: Dict[str, Any]={}):
		"""Calls the specified Trigger of another Area with the optionally given parameters.

		Args:
			target: The name of the target Area.
			trigger: The name of the Trigger.
			parameters: The parameters of the function call.
		"""
		pass

	@abstractmethod
	def scheduling(self, func: Callable) -> Callable:
		"""Registers this function as scheduler (only one). Intended to be used as decorator."""
		pass
	
	# ------------- Trigger processing functions -------------
	
	@abstractmethod
	def register(self, func: Callable) -> Callable:
		"""Registers the function as remote Trigger. Intended to be used as decorator."""
		pass
	
	@abstractmethod
	def comply(self, trigger: str, parameters: Dict[str, Any]={}):  # TODO: args/kwargs?
		"""Registers a request to call the Trigger function with the given parameters during the next round.

		This method should be induced by Areas higher in the hierarchy, to influence this Areas procedures.
		A certain trigger will only be called once in a single round.
		The parameters of the most recent request will be used.
		This method should not be called from within the Area.
		Autoloop requests are given preference.

		Args:
			trigger: The name of the trigger.
			parameters: The parameters of the function call.
		"""
		pass
	
	@abstractmethod
	def autoloop(self, trigger: str, parameters: Dict[str, Any]={}):  # TODO: args/kwargs?
		"""Registers a request to call the Trigger function with the given parameters during the next round.

		A certain trigger will only be called once in a single round.
		The parameters of the most recent request will be used.
		This method is intended to be called from within the Area.

		Args:
			trigger: The name of the trigger.
			parameters: The parameters of the function call.
		"""
		pass
	
	# ------------- Message processing functions -------------
	
	@abstractmethod
	def alteration(self, func: Callable=None, source: str=None, slot: str=None) -> Callable:
		"""Registers the function to be called in the next round after a optionally specified Subscriber message has arrived.

		Intended to be used as decorator.
		Only one function can be registered for a certain source/slot combination.

		Args:
			func: The function to be called.
			source: The optional name of the Area that sent the message.
			slot: The optional Slot from which the message was sent. A source must be given.
		"""
		pass
	
	@abstractmethod
	def process(self, source: str, message: dict, slot: str=None):
		"""Processes the given message. A message should originate from an Area lower in the hierarchy.

		This method should be induced by Areas lower in the hierarchy, to send their results.
		Data processing should be done by alteration functions.
		This method should not be called from within the Area.

		Args:
			source: The name of the Area that sent the message.
			message: The message to be processed.
			slot: The optional Slot from which the message was sent.
		"""
		pass
	
	@abstractmethod
	def alter(self, source: str=None, slot: str=None):
		"""Requests an alteration function to be called in the next round, even if no Subscriber message has arrived.
		
		A certain function will only be called once in a single round.
		This method is intended to be called from within the Area.

		Args:
			source: The source value the desired function was registered with.
			slot: The slot value the desired function was registered with.
		"""
		pass
	
	# ------------- Clock processing functions -------------
	
	@abstractmethod
	def activity(self, func: Callable) -> Callable:
		"""Registers the function to be called if there occurred input during this round.
		
		Only one function can be set.
		This method is intended to be used as a decorator.
		"""
		pass
	
	@abstractmethod
	def pulse(self, func: Callable) -> Callable:
		"""Registers the function to be called at the end of every round.
		
		Only one function can be set.
		This method is intended to be used as a decorator.
		"""
		pass
	
	@abstractmethod
	def proceed(self):
		"""Initiates the next round of this Areas procedures. Requested functions will be called.
		
		This method should be induced by a syncing Node, like a common Clock.
		This method should not be called from within the Area.
		"""
		pass



import json
import logging
from typing import Any, Dict

import zmq

from ..abcs import Node
from .utils import Socket, Context


logger = logging.getLogger(__name__)


class Publisher:
	def __init__(self, ip: str, port: int, host: str=None, context: Context=None):
		self._zmq = context or zmq.Context.instance()  # type: Context
		self.ip = ip  # type: str
		self.port = port  # type: int
		self.host = host  # type: Optional[str]
		self._publisher = None  # type: Socket
	
	def publish(self, message: dict, slot: str=None):
		if self._publisher is None:
			self._publisher = self._zmq.socket(zmq.PUB)
			self._publisher.bind('tcp://{}:{}'.format(self.host or self.ip, self.port))
		message = json.dumps(message)
		self._publisher.send_multipart([(slot or '').encode('utf-8'), message.encode('utf-8')])
		logger.debug('Message published at "%s"', slot)


class Pusher:
	def __init__(self, node: Node, context: Context):
		self._node = node  # type: Node
		self._zmq = context  # type: Context

	def trigger(self, target: str, trigger: str, parameters: Dict[str, Any]={}):
		ip, port, host = self._node.rslv(name=target, interface='rcv')
		pusher = self._zmq.socket(zmq.PUSH)  # type: Socket
		pusher.connect('tcp://{}:{}'.format(ip, port))
		parameters['trigger'] = trigger
		pusher.send_json(parameters)
		pusher.close()
		logger.debug('Trigger "%s" of "%s" called', trigger, target)



from typing import Any, Dict

import zmq

from .extensions import Publisher, Pusher
from .utils import Context
from ..abcs import Node
from ..utils import Interfaces


class Runner(Node):
	"""Abstract class to implement Pabiana Nodes with the ability to publish messages and call Triggers based on ZMQ."""
	def __init__(self, name: str, interfaces: Interfaces):
		super().__init__(name, interfaces)
		self._zmq = zmq.Context.instance()  # type: Context
		self.publish = Publisher(*self.rslv('pub'), self._zmq).publish
		self.trigger = Pusher(self, self._zmq).trigger

	def publish(self, message: dict, slot: str = None): pass

	def trigger(self, target: str, trigger: str, parameters: Dict[str, Any]={}): pass



from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

setup(
	name='Pabiana',
	version=open(path.join(here, 'VERSION.txt')).read().strip(),
	packages=find_packages(),
	install_requires=open(path.join(here, 'requirements.txt')).read().split('\n'),
	setup_requires=['pytest-runner >= 4.4'],
	tests_require=['pytest >= 4.4.1'],
	zip_safe=False,
	
	url='https://github.com/kankiri/pabiana',
	author='Alexander Schöberl',
	author_email='alexander.schoeberl@gmail.com',
	description='A minimalistic framework to build distributed cognitive applications based on ØMQ',
	long_description=open(path.join(here, 'README.md')).read(),
	long_description_content_type='text/markdown',
	keywords=['framework', 'cognitive', 'distributed', 'ØMQ'],
	license='MIT'
)
